//
//  This example uses the tranform_feedback extension to perform
//  recursive subdivision of geometry.
//
//  It uses two vertex buffer objects. The first vbo is initialized
//  with some simple geometry. This vbo is rendered with a geometry
//  program to create new geometry, and the result is stored to the
//  other vbo using transform feedback. The vbos are then swapped and
//  the process continues.
//
// Author: Simon Green
// Email: sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <stdio.h>
#include <map>

#include <GL/glew.h>
#include <GL/glut.h>

#include <Cg/cgGL.h>

#include <nvGlutManipulators.h>
#include <nvSDKPath.h>

////////////////////////////////////////////////////////////////////////////////
//
// Globals
//
////////////////////////////////////////////////////////////////////////////////

#define DEMO_PATH "src/transform_feedback_fractal/"
nv::SDKPath sdkPath;

using std::map;

nv::GlutExamine manipulator;

enum UIOption {
    OPTION_DISPLAY_WIREFRAME,
    OPTION_ENABLE_LIGHTING,
    OPTION_CONTINUOUS_SUBDIVISION,
    OPTION_ANIMATE,
    OPTION_COUNT
};
bool options[OPTION_COUNT];
map<char,UIOption> optionKeyMap;

GLuint vbo[2];
int current_buffer = 0;
const int max_buffer_verts = 1<<20;
GLuint query;
GLuint primitives_written = 1;
int verts_per_prim = 2;
GLenum render_prim = GL_LINES, input_prim = GL_LINES, feedback_prim = GL_LINES;
int prim_output_div = 1;

CGcontext cg_context; 
CGprofile cg_vprofile, cg_gprofile, cg_fprofile;
CGprogram passthru_vprog, xform_vprog;
CGprogram koch_gprog, terrain_gprog, branch_gprog, subdivide_gprog;
CGprogram terrain_display_gprog, terrain_calc_normal_gprog;
CGparameter terrain_rand_scale, terrain_rand_xform;

GLuint rand_tex;
float init_scale = 0.5;
float rand_scale;
float rand_xform_scale = 0.5;
float rand_xform_offset = 0.5;

void init_koch();
void init_terrain();
float frand();
GLuint create_rand_texture(GLenum target, GLenum format, int w, int h);

int mode = 2;
int step = 0;
int max_steps = 9;
int steps = 8;
int seed = 1973;

////////////////////////////////////////////////////////////////////////////////
//
// Functions
//
////////////////////////////////////////////////////////////////////////////////

//
//
//////////////////////////////////////////////////////////////////////
void init_opengl()
{
    glEnable(GL_DEPTH_TEST);
    glClearColor(0.2, 0.2, 0.2, 1.0);

    glewInit();

    if (!glewIsSupported(
        "GL_VERSION_2_0 "
        "GL_ARB_vertex_program "
        "GL_NV_gpu_program4 "
        "GL_NV_transform_feedback "
        "GL_ARB_texture_float "
        ))
    {
        printf("Unable to load extensions\n\nExiting...\n");
        exit(-1);
    }

    // create buffer objects
    glGenBuffers(2, vbo);
    glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
    glBufferData(GL_ARRAY_BUFFER, max_buffer_verts*4*sizeof(GLfloat), 0, GL_STATIC_DRAW);

    // load initial vertex data
    GLfloat v[] = {
        -1.0, 0.0, 0.0, 1.0,
        1.0, 0.0, 0.0, 1.0,
    };
    glBufferSubData(GL_ARRAY_BUFFER, 0, 2*4*sizeof(GLfloat), v);

    glBindBuffer(GL_ARRAY_BUFFER, vbo[1]);
    glBufferData(GL_ARRAY_BUFFER, max_buffer_verts*4*sizeof(GLfloat), 0, GL_STATIC_DRAW);

    // specify which attributes to store
    GLint attribs[] = { GL_POSITION, 4, 0 };
    glTransformFeedbackAttribsNV(1, attribs, GL_SEPARATE_ATTRIBS_NV);

    // create query objects
    glGenQueries(1, &query);

//    init_koch();
    init_terrain();
}

//
//
//////////////////////////////////////////////////////////////////////
void cgErrorCallback()
{
    CGerror lastError = cgGetError();
    if(lastError)
    {
        printf("%s\n", cgGetErrorString(lastError));
        printf("%s\n", cgGetLastListing(cg_context));
        exit(1);
    }
}

//
//
//////////////////////////////////////////////////////////////////////
void init_cg()
{
    // Create Cg Context
    cg_context = cgCreateContext();
    cgSetErrorCallback(cgErrorCallback);

    // load Cg programs
    cg_vprofile = cgGLGetLatestProfile(CG_GL_VERTEX);
    cg_gprofile = cgGLGetLatestProfile(CG_GL_GEOMETRY);
    cg_fprofile = cgGLGetLatestProfile(CG_GL_FRAGMENT);

    std::string resolved_path;

    if ( sdkPath.getFilePath( DEMO_PATH "shaders/subdivide.cg", resolved_path)) {
        xform_vprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_vprofile, "xform_vp", 0);
        cgGLLoadProgram(xform_vprog);

        passthru_vprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_vprofile, "passthru_vp", 0);
        cgGLLoadProgram(passthru_vprog);

        koch_gprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_gprofile, "koch_gp", 0);
        cgGLLoadProgram(koch_gprog);

        branch_gprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_gprofile, "branch_gp", 0);
        cgGLLoadProgram(branch_gprog);

        terrain_gprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_gprofile, "terrain_gp", 0);
        cgGLLoadProgram(terrain_gprog);
        terrain_rand_xform = cgGetNamedParameter(terrain_gprog, "rand_xform");
        terrain_rand_scale = cgGetNamedParameter(terrain_gprog, "rand_scale");

        terrain_calc_normal_gprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_gprofile, "terrain_calc_normal_gp", 0);
        cgGLLoadProgram(terrain_calc_normal_gprog);

        terrain_display_gprog = cgCreateProgramFromFile(cg_context, CG_SOURCE, resolved_path.c_str(), cg_gprofile, "terrain_display_gp", 0);
        cgGLLoadProgram(terrain_display_gprog);
    }
    else {
        fprintf( stderr, "Failed to find shader file '%s'\n'", DEMO_PATH "shaders/subdivide.cg");
    }

    rand_tex = create_rand_texture(GL_TEXTURE_2D, GL_RGBA16F_ARB, 256, 256);
}

//
//
//////////////////////////////////////////////////////////////////////
float frand()
{
    return rand() / (float) RAND_MAX;
}

//
// create a texture containing random numbers
// 
//////////////////////////////////////////////////////////////////////
GLuint create_rand_texture(GLenum target, GLenum format, int w, int h)
{
  float *data = new float [w*h*4];
  float *ptr = data;
  for(int i=0; i<w*h; i++) {
    *ptr++ = frand()*2.0-1.0;
    *ptr++ = frand()*2.0-1.0;
    *ptr++ = frand()*2.0-1.0;
    *ptr++ = frand()*2.0-1.0;
  }

  GLuint texid;
  glGenTextures(1, &texid);
  glBindTexture(target, texid);

  glTexParameteri(target, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(target, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(target, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(target, GL_TEXTURE_WRAP_T, GL_REPEAT);

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glTexImage2D(target, 0, format, w, h, 0, GL_RGBA, GL_FLOAT, data);

  return texid;
}

//
// load initial vertex data for koch snowflake pattern
//
//////////////////////////////////////////////////////////////////////
void init_koch()
{
    glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
    GLfloat v[] = {
        1.0, -1.0, 0.0, 1.0,
        -1.0, -1.0, 0.0, 1.0,

        -1.0, -1.0, 0.0, 1.0,
        0.0, 1.0, 0.0, 1.0,

        0.0, 1.0, 0.0, 1.0,
        1.0, -1.0, 0.0, 1.0,
    };
    glBufferSubData(GL_ARRAY_BUFFER, 0, 6*4*sizeof(GLfloat), v);

    current_buffer = 0;
    verts_per_prim = 2;
    prim_output_div = 1;
    input_prim = render_prim = feedback_prim = GL_LINES;
    primitives_written = 3;
    subdivide_gprog = koch_gprog;
}

//
// load initial vertex data for terrain
//
//////////////////////////////////////////////////////////////////////
void init_terrain()
{
    GLfloat h[4];
    for(int i=0; i<4; i++)
        h[i] = frand();

    GLfloat v[] = {
        -1.0, h[0], -1.0, 1.0,
        1.0, h[1], -1.0, 1.0,
        1.0, h[2], 1.0, 1.0,
        -1.0, h[3], 1.0, 1.0,
    };

    glBindBuffer(GL_ARRAY_BUFFER, vbo[0]);
    glBufferSubData(GL_ARRAY_BUFFER, 0, 4*4*sizeof(GLfloat), v);

    current_buffer = 0;
    verts_per_prim = 4;
    prim_output_div = 4;
    input_prim = GL_LINES_ADJACENCY_EXT;
    render_prim = GL_QUADS;
    feedback_prim = GL_POINTS;
    primitives_written = 4;
    subdivide_gprog = terrain_gprog;
    rand_scale = init_scale;
}

//
//
//////////////////////////////////////////////////////////////////////
void init()
{
    switch(mode) {
    case 0:
        init_koch();
        options[OPTION_ENABLE_LIGHTING] = false;
        break;
    case 1:
        init_koch();
        subdivide_gprog = branch_gprog;
        options[OPTION_ENABLE_LIGHTING] = false;
        break;
    case 2:
        init_terrain();
        options[OPTION_ENABLE_LIGHTING] = true;
        break;
    }
}

//
// display geometry stored in buffer object
//
//////////////////////////////////////////////////////////////////////
void draw_buffer_geom(GLuint buffer, GLenum prim, int verts)
{
    glBindBuffer(GL_ARRAY_BUFFER, buffer);
    glVertexAttribPointerARB(0, 4, GL_FLOAT, GL_FALSE, 0, 0);
    glEnableVertexAttribArrayARB(0);
    glDrawArrays(prim, 0, verts);
    glDisableVertexAttribArrayARB(0);
}

//
// render object, storing positions to buffer object
//
//////////////////////////////////////////////////////////////////////
void do_feedback()
{
    // set base for transform feedback
    glBindBufferOffsetNV(GL_TRANSFORM_FEEDBACK_BUFFER_NV, 0, vbo[1-current_buffer], 0);

    glBeginTransformFeedbackNV(feedback_prim);
    glEnable(GL_RASTERIZER_DISCARD_NV);    // disable rasterization

    glBeginQuery(GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN_NV, query);

    cgGLBindProgram(passthru_vprog);
    cgGLEnableProfile(cg_vprofile);

    cgGLBindProgram(subdivide_gprog);
    cgGLEnableProfile(cg_gprofile);
    cgGLSetParameter1f(terrain_rand_scale, rand_scale);
    cgGLSetParameter2f(terrain_rand_xform, rand_xform_scale, rand_xform_offset);

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, rand_tex);

    draw_buffer_geom(vbo[current_buffer], input_prim, primitives_written*verts_per_prim/prim_output_div);

    glEndQuery(GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN_NV);

    cgGLDisableProfile(cg_vprofile);
    cgGLDisableProfile(cg_gprofile);
    glDisable(GL_RASTERIZER_DISCARD_NV);
    glEndTransformFeedbackNV();

    // read back query results
    glGetQueryObjectuiv(query, GL_QUERY_RESULT, &primitives_written);

    current_buffer = 1 - current_buffer;
    rand_scale *= 0.5;
}


//
//
//////////////////////////////////////////////////////////////////////
void display() {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    if ( options[OPTION_CONTINUOUS_SUBDIVISION]) {
        // continuous subdivision
        srand(seed);
        init();
        for(int i=0; i<steps; i++) {
            do_feedback();
        }
    }

    // wireframe control
    glPolygonMode( GL_FRONT_AND_BACK, options[OPTION_DISPLAY_WIREFRAME] ? GL_LINE : GL_FILL);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    manipulator.applyTransform();

    cgGLBindProgram(xform_vprog);    // note - need vertex program for generic attribs to work!
    cgGLEnableProfile(cg_vprofile);
    glColor3f(1.0, 1.0, 1.0);
    if ( options[OPTION_ENABLE_LIGHTING]) {
        cgGLBindProgram(terrain_display_gprog);
        cgGLEnableProfile(cg_gprofile);
        glShadeModel(GL_FLAT);
        draw_buffer_geom(vbo[current_buffer], input_prim, primitives_written*verts_per_prim/prim_output_div);
    } else {
        draw_buffer_geom(vbo[current_buffer], render_prim, primitives_written*verts_per_prim/prim_output_div);
    }

    cgGLDisableProfile(cg_vprofile);
    cgGLDisableProfile(cg_gprofile);

    glutSwapBuffers();
    glutReportErrors();
}

//
//
//////////////////////////////////////////////////////////////////////
void idle() {
    if (options[OPTION_ANIMATE]) {
        manipulator.idle();
        glutPostRedisplay();
    }
}

//
//
//////////////////////////////////////////////////////////////////////
void key(unsigned char k, int x, int y) {
    k = tolower(k);

    if (optionKeyMap.find(k) != optionKeyMap.end()) {
        // toggle option
        options[optionKeyMap[k]] = !options[optionKeyMap[k]];
        if (optionKeyMap[k] == OPTION_CONTINUOUS_SUBDIVISION && options[optionKeyMap[k]] == false) {
            init();
            step = 0;
        }
    }

    switch(k) {
        case 27:
        case 'q':
            exit(0);

        case 's':
            if (step < max_steps) {
                do_feedback();
                step++;
            }
            break;

        case 'd':
            printf("primitives = %d\n", primitives_written);
            break;

        case '1':
        case '2':
        case '3':
            mode = k - '1';
            init();
            step = 0;
            break;

        case 'r':
            step=0;
            init();
            break;

        case '+':
        case '=':
            if (steps < max_steps) steps++;
            break;

        case '-':
            if (steps > 0) steps--;
            break;

        case ']':
            rand_xform_scale += 0.001;
            break;
        case '[':
            rand_xform_scale -= 0.001;
            break;
    }
    
    glutPostRedisplay();
}

//
//
//////////////////////////////////////////////////////////////////////
void resize(int w, int h)
{
    glViewport(0, 0, w, h);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
    gluPerspective(60.0, (GLfloat)w/(GLfloat)h, 0.01, 10.0);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    manipulator.reshape(w, h);
}

//
//
//////////////////////////////////////////////////////////////////////
void mouse(int button, int state, int x, int y)
{
    manipulator.mouse(button, state, x, y);
}

//
//
//////////////////////////////////////////////////////////////////////
void motion(int x, int y)
{
    manipulator.motion(x, y);
}

//
//
//////////////////////////////////////////////////////////////////////
void main_menu(int i)
{
  key((unsigned char) i, 0, 0);
}

//
//
//////////////////////////////////////////////////////////////////////
void init_menus()
{
    glutCreateMenu(main_menu);
    glutAddMenuEntry("Kock curve [1]", '1');
    glutAddMenuEntry("Branch [2]", '2');
    glutAddMenuEntry("Terrain [3]", '3');
    glutAddMenuEntry("Iterate subdivision [ ]", ' ');
    glutAddMenuEntry("Reset [r]", 'r');
    glutAddMenuEntry("Toggle lighting [l]", 'l');
    glutAddMenuEntry("Toggle continuous subdivision [c]", 'c');
    glutAddMenuEntry("Dump feedback buffer [d]", 'd');
    glutAddMenuEntry("Quit (esc)", '\033');
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

//
//
//////////////////////////////////////////////////////////////////////
int main(int argc, char **argv) {
    glutInit(&argc, argv);
    glutInitWindowSize(512, 512);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);
    glutCreateWindow("transform_feedback_fractal");

    init_cg();
    init_opengl();
    init_menus();

    manipulator.setDollyActivate( GLUT_LEFT_BUTTON, GLUT_ACTIVE_CTRL);
    manipulator.setPanActivate( GLUT_LEFT_BUTTON, GLUT_ACTIVE_SHIFT);
    manipulator.setDollyPosition( -3.0f);

    glutDisplayFunc(display);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutIdleFunc(idle);
    glutKeyboardFunc(key);
    glutReshapeFunc(resize);

    //configure the options
    optionKeyMap['w'] = OPTION_DISPLAY_WIREFRAME;
    options[OPTION_DISPLAY_WIREFRAME] = false;

    optionKeyMap['l'] = OPTION_ENABLE_LIGHTING;
    options[OPTION_ENABLE_LIGHTING] = true;
    
    optionKeyMap[' '] = OPTION_ANIMATE;
    options[OPTION_ANIMATE] = true;

    optionKeyMap['c'] = OPTION_CONTINUOUS_SUBDIVISION;
    options[OPTION_CONTINUOUS_SUBDIVISION] = true;

    //print the help info
    printf( "transform_feedback_fractal - sample showing the usage of GL_NV_transform_feedback\n");
    printf( "  Commands:\n");
    printf( "    q / [ESC] - Quit the application\n");
    printf( "    [SPACE]   - Toggle continuous animation\n");
    printf( "    l         - Toggle lighting\n");
    printf( "    w         - Toggle displaying wireframe\n");
    printf( "    c         - Toggle continuous subdivision\n");
    printf( "    +         - Increase number of iterations of feedback\n");
    printf( "    -         - Decrease number of iterations of feedback\n");
    printf( "    ]         - Increase random transform scale\n");
    printf( "    [         - Decrease random transform scale\n");
    printf( "    r         - Reset values\n");
    printf( "    s         - Step through feedback iteractions (only meaningful if continuous subdivision is off)\n");
    printf( "    [mouse]   - Left button - rotate, Left + ctrl - dolly, Left + shift - pan\n\n");

    glutMainLoop();

    return 0;
}